
#include <cmath>
#include <iostream>
#include <vector>
#include <algorithm>
#include <limits>

using std::vector;

using namespace std;

vector<float> get_mean_point_distance(vector<int> & x, vector<int> & y, vector<int> & z);